//========================================================================
//=
//=     *** Kemper LED Lamp ***					10/29/2009
//=
//= This source code trasforms a 12F609 micro into a KLL ASCI (Kemper LED
//= Lamp - Application Specific Integrated Circuit). The chip becomes a 
//= slave node on a TTL RS232 bus operating at 9600 baud / half duplex.
//= Each node should have a unique node address on the bus. Otherwise, the 
//= two nodes will fight for the bus. Once flashed, the KLL node will 
//= respond to almost a dozen commands. The commands alow for fine PWM 
//= control of the four output channels on the 12F609 chip. Use the commands,
//= the KLL will smothly ramp, hold, and decay the output channels.
//=
//========================================================================

/*
 Powerhouse Electronics 2010, Some Rights Reserved 
 Kemper LED Lamp 2010, Some Rights Reserved 
    Except where otherwise noted, this work is licensed under 
    This document is licensed under the Attribution-NonCommercial-ShareAlike 3.0 
    Unported license, available at http://creativecommons.org/licenses/by-nc-sa/3.0/.

You are free:
   * to Share  to copy, distribute and transmit the work
   * to Remix  to adapt the work
   * not for Commercial use unless presission is given.

Under the following conditions:
   * Attribution. You must attribute the work in the manner specified by the author 
   or licensor (but not in any way that suggests that they endorse you or your use 
   of the work).
   * Share Alike. If you alter, transform, or build upon this work, you may distribute 
   the resulting work only under the same, similar or a compatible license.
   * For any reuse or distribution, you must make clear to others the license terms 
   of this work. The best way to do this is with a link to this web page.
   * Any of the above conditions can be waived if you get permission from the copyright
   holder.
   * Nothing in this license impairs or restricts the author's moral rights.
*/


#include "ASIC.h"

// ----------------------------------------------------------------------------

// Setup a receive buffer for comm characters. 
#define MAX_PACKET_LEN	8
byte buff[ MAX_PACKET_LEN ];

// Keep a threshold for each LED. When the timer rolls by the threshold ServiceLEDs
// will change the output pin. 
byte iBlue, iGreen, iRed, iWhite;

// These are the target values for each color.
byte iBlueTar, iGreenTar, iRedTar, iWhiteTar;

// Rate at which the LEDs are ramped up and down.
byte iRBlue,    iDBlue;
byte iRGreen,   iDGreen;
byte iRRed,     iDRed;
byte iRWhite,   iDWhite;

#define ALT_NODES	3
byte iAltNode[ALT_NODES];   // Second address.
byte ind;                   // Index for receive buffer.
byte len;                   // Expected lenght of packet.
byte chr, ii,u8Mark;        // IRQ variables.

// Keep the declaration of booleans all together so the compiler can pack them
// into one byte.
bool fStretch, fNodeAck, fTmp;

byte iDwell, iDwellTmr, iPulLED;

// Define how many expected characters we need to
// see for each command type. Make sure array is at least 16 elements
// long. So, out nibble can't index off the end.
const byte iPackLen[] = {
    3,  //  RESET			        0
    3,  //  SET_PULLUP			    1   
    4,  //  SET_LED_ON_OFF		    2
    5,  //  QUICK_SET_LED_LEVEL		3
    5,  //  SET_LED_RAMP_DECAY		4
    4,  //  SET_LED_LEVEL		    5
    3,  //  START_TURBO_PING		6
    3,  //  RESET_PWM_TIMER		    7
    4,  //  SET_ACK_MODE		    8
    6,  //  SET_2ND_ADDR		    9
    4,  //  PULSE_LED			    10
    5,  //  Null			        11
    5,  //  Null			        12
    5,  //  Null			        13
    5,  //  Null			        14
    5   //  Extend Cmd			    15
};




//========================================================================
// Main Loop...
//========================================================================
void main() {
    byte cmd, i;

    init();                 // Initialize everything.

    while ( 1 ) {

        vServiceLEDs();     // No mater what, keep the LEDs updating!

        // On timer0 overflow (approx. every 16.5ms) ramp LED(s)...
        // This is the lowest priority task in the system. Nobody will
        // notice a ramp that takes just a bit longer.
        if ( T0IF ) {
            T0IF = false;     // Clear timer0 overflow flag.

            vServiceLEDs();
            vRampLED( &iRed, &iRedTar, iRRed, iDRed, 0x08 );

            vServiceLEDs();
            vRampLED( &iGreen, &iGreenTar, iRGreen, iDGreen, 0x04 );

            vServiceLEDs();
            vRampLED( &iBlue, &iBlueTar, iRBlue, iDBlue, 0x02 );

            vServiceLEDs();
            vRampLED( &iWhite, &iWhiteTar, iRWhite, iDWhite, 0x01 );

            fStretch = false;
        }

        // Jump the timer count back - this has the effect of stretching the
        // timer from 256 counts to 260. This makes the roll over every 60Hz 
        // instead of 61Hz. This becomes important for video cameras. The human
        // eye won't see an adjustment around 0xA0 counts. The eye will see the
        // adjustment if made around zero!
        if ( (TMR0 & 0xA0) && (fStretch == false) ) {
            TMR0 -= 4;
            fStretch = true;      // Flag that stretch is complete.
        }

        // After the second buffer byte, get the expected packet length stored away.
        if ( ind > 1 ) {
            cmd = buff[1] & 0x0F;     // Limit cmd to a nibble!
            len = iPackLen[ cmd ];        // Lookup the expected packet length.
        }

        // On a new packect...
        if ( ind >= len ) {
            disable_interrupts( GLOBAL );

            if ( fChecksum( len ) ) {
                if ( buff[0] == NODE_ADDR ) {
                    // Ack a good packet if it was sent directly to this node. 
                    if ( fNodeAck ) Tx( NODE_ADDR );
                    vParsePacket();
                }
                else {
                    for ( i=0 ; i < ALT_NODES ; i++ ) {
                        if ( buff[0] == iAltNode[i] ) {
                            // Alt node addresses don't get acked.
                            vParsePacket();
                            break;
                        }
                    }
                }
            }
            // On a bad checksum, but packet for this node, send a Nack.
            else if ( fNodeAck && (buff[0] == NODE_ADDR) ) {
                Tx( NACK );
            }

            // After a packet is seen, the node needs to ignore any other nodes that 
            // might be responding on the bus. So, after a packet, go deaf for several
            // milliseconds.
            set_timer1( 0 );
            while ( get_timer1() < m2ms ) {       // 2ms timeout.
                vServiceLEDs();
            }

            set_timer1( 0 );                    // Reset the timeout timer.
            len = MAX_PACKET_LEN;               // Just make sure ind and ind don't match.
            ind = 0;                            // Clearing index allows a new packet to be received.
            while ( INTF )                      // Make sure the irq flag is cleared!
                INTF = 0;
            enable_interrupts(GLOBAL);
        }

        // A timeout (60ms) occured while waiting for a full packet.
        if ( ind && (get_timer1() > m60ms) ) {
            ind = 0;                            // Reset packet byte index.
            len = MAX_PACKET_LEN;
            set_timer1( 0 );                    // Clear timeout timer.
        }
    } 
}

//========================================================================
void vParsePacket( void ) {
    byte i;

    // Parse on just the lowwer nibble in the command byte.
    switch ( buff[1] & 0x0F ) {     // Mask off (dump) upper nibble from byte.

    // 0						
        case RESET:                        
            init();                      
            break;                      

    // 1
        case SET_PULLUP:
            if ( buff[1] & 0xF0 )   WPU = PULLUP;   // Turn on weak pullup for COMM pin.
            else                    WPU = 0x00;
            break;

    // 2						 
        case SET_LED_ON_OFF:
            i = bLevels[ buff[2] >> 4 ]; // Use upper nibble as index to level.
            if ( buff[1] & mRED ) if ( buff[2] & 0x08 ) iRedTar = i;
                else iRedTar = 0;
            if ( buff[1] & mGRN ) if ( buff[2] & 0x04 ) iGreenTar = i;
                else iGreenTar = 0;
            if ( buff[1] & mBLU ) if ( buff[2] & 0x02 ) iBlueTar = i;
                else iBlueTar = 0;
            if ( buff[1] & mWHT ) if ( buff[2] & 0x01 ) iWhiteTar = i;
                else iWhiteTar = 0;
            break;                      


    // 3
        case QUICK_SET_LED_LEVEL:
            // Bomb out the compiler if the array is not defined to the right size.
            #if ( sizeof(bLevels) != 16 )
               #error "bLevels must be 16 bytes in length!"
            #endif	

            if ( buff[1] & mRED ) iRedTar = bLevels[ buff[2] >> 4 ];
            if ( buff[1] & mGRN ) iGreenTar = bLevels[ buff[2] & 0x0F ];
            if ( buff[1] & mBLU ) iBlueTar = bLevels[ buff[3] >> 4 ];
            if ( buff[1] & mWHT ) iWhiteTar = bLevels[ buff[3] & 0x0F ];
            break;

    // 4
        case SET_LED_RAMP_DECAY:
            if ( buff[1] & mRED ) {
                iRRed = buff[2]; iDRed = buff[3];
            }
            if ( buff[1] & mGRN ) {
                iRGreen = buff[2]; iDGreen = buff[3];
            }
            if ( buff[1] & mBLU ) {
                iRBlue = buff[2]; iDBlue = buff[3];
            }
            if ( buff[1] & mWHT ) {
                iRWhite = buff[2]; iDWhite = buff[3];
            }
            break;

    // 5						
        case SET_LED_LEVEL:
            if ( buff[1] & mRED ) iRedTar = buff[2];
            if ( buff[1] & mGRN ) iGreenTar = buff[2];
            if ( buff[1] & mBLU ) iBlueTar = buff[2];
            if ( buff[1] & mWHT ) iWhiteTar = buff[2];
            break;

    // 6
        case START_TURBO_PING:
            // Create a dwell of 2ms * NODE_ADDRESS
            i = 0;  
            while ( ++i != 0 ) {                                // Wait for ALL the nodes to report.
                set_timer1( 0 );                                // Clear timer to zero.
                while ( get_timer1() < m2ms ) vServiceLEDs();   // Wait 2ms (but, keep LEDs updating).
                if ( i == NODE_ADDR ) Tx( NODE_ADDR );          // Put our node address on the bus.
            }
            break;

    // 7
        case RESET_PWM_TIMER:
            set_timer0( 0 );    // Reset LED timer (used to sync timer with other nodes).
            break;

    // 8
        case SET_ACK_RESPONSE:
            // Mask off the command nibble.
            if ( buff[1] & 0xF0 )   fNodeAck = true;
            else                    fNodeAck = false;
            iDwell = buff[2];
            break;


    // 9
        case SET_2ND_ADDR:
            iAltNode[0] = buff[2];
            iAltNode[1] = buff[3];
            iAltNode[2] = buff[4];
            #if ALT_NODES != 3
               // Verify how you have ALT_NODES defined!
               #error "Wrong number of ALT_NODES"
            #endif
            break;

        case PULSE_LED:
            iDwellTmr = iDwell;
            if ( buff[1] & mRED ) {
                iRedTar = buff[2]; iPulLED |= 0x08;
            }
            if ( buff[1] & mGRN ) {
                iGreenTar = buff[2]; iPulLED |= 0x04;
            }
            if ( buff[1] & mBLU ) {
                iBlueTar = buff[2]; iPulLED |= 0x02;
            }
            if ( buff[1] & mWHT ) {
                iWhiteTar = buff[2]; iPulLED |= 0x01;
            }
            iDwellTmr = iDwell; 
            break;

        default:
            break;
    }

}


// Given some pointers and rate rates, this function makes one step
// change adjustment to the output threshold. This function should be 
// called every 16.384ms. Example, if ramp equaled one, and thresh started
// at zero, it would take over 4 seconds to ramp the threshold from zero 
// to 0xFF at 16.384 ms per step.
//========================================================================
void vRampLED( byte *thresh, byte *tar, ramp, decay, byte iIndex ) {

    // If threshold equals target then do nothing.

    // If threshold is below target then use "ramp" to bring up the threshold.
    if ( *thresh < *tar ) {
        *thresh += ramp;                               
        // Use the status bit overflow to check if we rolled over.
        if ( MATHOF || ( *thresh >= *tar) ) *thresh = *tar;
    }

    // If threshold is above the target then use "decay" to bring down the threshold.
    if ( *thresh > *tar ) {
        *thresh -= decay;                               
        // Use the status bit overflow to check if we went negative.
        if ( ~MATHOF || ( *thresh <= *tar) ) *thresh = *tar;
    }

    // If pulsing an led is true...
    if ( iPulLED & iIndex ) {
        // Once the led is at the target level start counting down the timer.
        if ( (*thresh == *tar) && iDwellTmr ) --iDwellTmr;
        // Once the timer times out, jump the target to zero and turn off the pulse bit.
        if ( iDwellTmr == 0 ) {
            *tar = 0;
            iPulLED &= ~iIndex;
        }
    }
}


// Compare the LED threshold to the current value in Timer0. On a threshold
// greater than Timer0, then set the output pin high. Else, set the pin 
// low. Note, if the threshold is zero then the output is always low 
// regardless of the timer. Also, in the same vain, a threshold of 0xFF 
// will always have the output pin high. This makes me happy when I look
// at my logic analyzer.
// Note, timer0 is counting at 64uS per tick. Therefore, this function must 
// be called more often than that if the LEDs are to look nice.
//
// Call this function often! 
// Also, call this function from inside interrupts (it's OK and safe)!
//========================================================================
void vServiceLEDs( void ) {

    if ( iBlue == 0 )           pBLU = true;
    else if ( iBlue >= TMR0 )   pBLU = false;
    else                        pBLU = true;

    if ( iGreen == 0 )          pGRN = true;
    else if ( iGreen >= TMR0 )  pGRN = false;
    else                        pGRN = true;

    if ( iRed == 0 )            pRED = true;
    else if ( iRed >= TMR0 )    pRED = false;
    else                        pRED = true;

    if ( iWhite == 0 )          pWHT = true;
    else if ( iWhite >= TMR0 )  pWHT = false;
    else                        pWHT = true;
}

//========================================================================
//void Tx( byte ch ) {
//  pCOMM = true;				// Turn A2 into an output pin. Default = High Level
//  SET_TRIS_A( 0b001000);		// 0=Output -> Bit Order: A5 A4 - A3 A2 A1 A0
//  #use rs232( baud=9600, xmit=PIN_A2 )
//  printf( "%c", ch );
//  SET_TRIS_A( 0b001100);		// Set A2 back to an input pin.
//}


//==============================================================================

void Tx( byte chOut ) {
    byte tmrMark, bit;
    // Timer1 tics at 4us. 
    // Each 9600 buad tick is 104.2us.
    set_timer1( 0 );          // Sync timer to start of comm event.
    pDDR_COMM = OUT;          // Set comm pin as output.
    pComm = 0;                // Start bit is low.
    tmrMark = 26;             // Start Bit Time
    bit = 0x01;               // Setup for the For loop.
    while ( get_timer1() < tmrMark ); // Wait for start bit.

    while ( bit ) {           // Loop through all eight bits.

        if ( chOut & bit )  pComm = 1;  // Send a bit if present in character.
        else                pComm = 0;

        bit <<= 1;              // Move mask bit left one click (0 shifted in).
        tmrMark += 26;          // Mark where we expect each new bit to start.
        while ( get_timer1() < tmrMark );   // Wait for tmr to catch up to our marker.
    }

    pComm = 1;                // Stop bit is high.
    pDDR_COMM = IN;           // Set comm pin as input.

    set_timer1( 0 );          // ReSync timer (so it does not roll over!).
    tmrMark = 15;
    while ( get_timer1() < tmrMark ); // Wait - last bit time (aligns  next character).
}



// Perform a simple addition of all the bytes in the receive packet. If the
// sum bytes equals the last byte (the checksum) then the checksum is good.
//========================================================================
bool fChecksum( byte len ) {
    byte i, cs;

  // Must have at least a byte and a checksum.
    if ( len < 2 ) return false;

    cs = 0;
    for ( i=0 ; i < (len-1) ; i++ ) {
        cs += buff[i];
        vServiceLEDs();
    }

    vServiceLEDs();
    if ( cs == buff[ len-1 ] )
        return true;
    return false;
}


// Initialize everything! Called after powerup and by reset command.
//========================================================================
void init( void ) {
    byte i;

    disable_interrupts( GLOBAL );// Disable global interrupts.

    ANSEL = 0x00;           // All digital - No Analog Pins.
    GPIO = 0b110111;        // Set all IO pins at once - LEDs are active low.
    SET_TRIS_A( 0b001100);  // 0=Output -> Bit Order: A5 A4 - A3 A2 A1 A0

    // Enable global weak pullups (active low). Only pin this affects 
    // is GP2 (the comm pin). All other pins are outputs. GP3/MCLR is ignored.
    GPPU = 0;                 
    WPU = PULLUP;           // Default weak pull-up on pCOMM.
    IOC = 0;                // Interrupt on change can't be trusted!

    // 500nS / Instruction
    // 1 / ((8Mhz Clock / 4 Cycles/Instruction) / 128 Prescaler) = 64uS / Tick
    setup_timer_0( RTCC_INTERNAL | RTCC_DIV_128 );    // Rollover = 16.384ms

    // 1 / ((8Mhz Clock / 4 Cycles/Instruction) / 8 Prescaler) = 4uS / Tick
    setup_timer_1( T1_INTERNAL | T1_DIV_BY_8 );

    // Init all the threshholds and targets.
    iRed = iGreen = iBlue = iWhite = 0;
    iRedTar = iGreenTar = iBlueTar = iWhiteTar = 0;

    // Init ramp rates to a nice rate.
    iRRed = iRGreen = iRBlue = iRWhite = 0x15;    // Ramp up rates.
    iDRed = iDGreen = iDBlue = iDWhite = 0x15;    // Decay down rates.

    iDwell = 0x10; iDwellTmr = 0; iPulLED = 0;

    //for ( i=0 ; i < ALT_NODES ; i++ )
    //  iAltNode[i] = 0xFF;
    iAltNode[0] = 0xFF;
    iAltNode[1] = 0xFF;
    iAltNode[2] = 0xFF;


    fNodeAck = true;                // Default: node sends back acks.

    // Clear buffer and index regs.
    for ( ii=0 ; ii < MAX_PACKET_LEN ; ++ii ) buff[ii] = 0;
    ind = 0; len = MAX_PACKET_LEN;
    set_timer1( 0 );                // Clear timeout timer.

    GPIF = 0;                       // Clear the I/O pin changed irq flag.
    INTF = 0;                       // Clear the INT request flag.
    PIE1 = 0x00;                    // No other interrupts allowed!

    ext_int_edge( H_TO_L );         // Setup interrupt as high to low.
    enable_interrupts( INT_EXT );   // Interrupt on GP2
    enable_interrupts( GLOBAL );
}


//========================================================================
// Comm IRQ:
// 
// This IRQ function receives new a new characters. The character is appended
// to the end of the receive buffer. While receiving the character, the LED 
// outputs are also serviced. This is super important or glitches will be 
// visible on the LEDs.
// 
// Timer1 is used exclusively to time an incoming characters. At the start of 
// each new character the timer is reset back to zero. 
//========================================================================
#INT_EXT
void vGetCharISR( void ) {

    set_timer1( 0 );          // Sync timer to strat of comm event.

    // Note, there is a constant 20us of latency on the interrupt.
    u8Mark = 33;              // Wait for 1 1/2 bits.

    //GP5 = false;
    vServiceLEDs();
    while ( get_timer1() < u8Mark );  // Wait for first 1 1/2 bit.

    // Capture first bit.
    chr = pCOMM;
    //GP5 = true;

    for ( ii=1 ; ii < 8 ; ii++ ) {    // Get seven more bits.
        u8Mark += 25;           // Mark where we expect each new bit to start.
        //GP5 = false;
        vServiceLEDs();         // Service the LEDs before we go into the while loop.
        while ( get_timer1() < u8Mark );    // Wait for TMR0 to catch up to our marker.
        fTmp = pCOMM;           // Latch I/O pin into boolean.
        //GP5 = true;
        chr |= fTmp << ii;          // Stuff new bit into position.
    }

    // Append new character to our buff buffer.
    if ( ind >= MAX_PACKET_LEN ) ind = MAX_PACKET_LEN - 1;
    buff[ ind ] = chr;
    ind += 1;

    u8Mark += 25;
    vServiceLEDs();
    while ( get_timer1() < u8Mark );  // Wait for the last bit time (aligns  next character).

    //GP5 = false;
    //GP5 = true;
}
